import json

from collections import defaultdict

import requests
import logging

from es import IndexReport

logger = logging.getLogger("verify_schema")


def verify_template(args, report, auth):
    """
    Loads a template from ES and compares it against the expected template from source
    
    :param args: program args
    :param report: report to fill
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :return: IndexReport
    """
    index = report.index
    logger.debug("Verifying '{index_name}' on {es_url}/_template/template-{index_name}".format(**index))
    response = requests.get("{es_url}/_template/template-{index_name}".format(**index), auth=auth, verify=False)

    indices = get_indices_for_template(index, auth)
    for index_name in indices:
        index_report = IndexReport(parent=report, name_override=index_name, is_template=False)
        index_schema = requests.get("{es_url}/{index}".format(index=index_name, **index), auth=auth, verify=False)
        logger.info(json.dumps(index_schema.json(), indent=2))
        verify_schema(args, list(index_schema.json().values())[0], index_report)

    if response.status_code == 404:
        report.is_missing = True
        return report
    response.raise_for_status()
    (present_name, actual) = list(response.json().items())[0]
    return verify_schema(args, actual, report)


def get_indices_for_template(index, auth):
    """
    Gets the indices created from a given index template
    
    :param index: The template to find indices for (not IndexReport)
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :return: string names of the indices
    """
    pattern = index.get("template_pattern", index["index_name"])

    # templates can specify their own wildcard format but if they don't, assume it's just a trailing wildcard
    if "*" not in pattern:
        pattern += "*"
    response = requests.get("{es_url}/{pattern}/_aliases".format(pattern=pattern, **index), auth=auth, verify=False)
    keys = list(response.json().keys())
    return keys


def verify_index(args, report, auth):
    """
    Loads an index from ES and compares it against the expected index from source
    
    :param args: program args
    :param report: report to fill
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :return: IndexReport
    """
    index = report.index
    logger.info("Verifying '{index_name}' on {es_url}/a_{index_name}".format(**index))

    response = requests.get("{es_url}/a_{index_name}".format(**index), auth=auth, verify=False)
    if response.status_code == 404:
        response = requests.get("{es_url}/{index_name}".format(**index), auth=auth, verify=False)
        if response.status_code == 404:
            logger.info("index '{index_name} is missing!".format(**index))
            report.is_missing = True
            return report
        else:
            logger.info("index '{index_name}' is missing its alias 'a_{index_name}'".format(**index))
    response.raise_for_status()

    (present_name, actual) = list(response.json().items())[0]
    return verify_schema(args, actual, report)


def verify_schema(args, actual, report):
    """
    Verifies the schema of an index (index or index template)
    
    :param args: program args
    :param actual: index, as reported by ES
    :param report: report to fill 
    :return: IndexReport
    """
    index = report.index
    report.actual_mappings = actual.get("mappings", {})
    report.actual_settings = actual.get("settings", {})

    if "mappings" in index and args.verify_mappings:
        if report.actual_mappings != index["mappings"]:
            key = list(index["mappings"].keys())[0]

            if key not in report.actual_mappings:
                if len(report.actual_mappings.keys()) == 0:
                    logger.info("index '{index_name}' has an alias on {es_url} but no mappings".format(**index))
                else:
                    logger.info("index '{index_name}' has an incorrect name configured.  "
                                "Expected: {expected}, found: {found}"
                                .format(expected=key, found=list(report.actual_mappings.keys())[0], **index))
                report.is_missing = True
            else:
                expected_props = index["mappings"][key]["properties"]
                actual_props = report.actual_mappings[key]["properties"]
                deep_compare_mappings(report, index, expected_props, actual_props)

    if "settings" in index and args.verify_settings:
        if "index" in report.actual_settings:
            if report.actual_settings["index"] != index["settings"]:
                deep_compare_settings(report, index, index["settings"], report.actual_settings["index"])
        else:
            fake_index = convert_dotted_keys_to_nested_dictionary(actual)
            deep_compare_settings(report, index, index["settings"], fake_index.get("index", {}))

    return report


def convert_dotted_keys_to_nested_dictionary(given_settings):
    """
    Sometimes elasticsearch returns the index settings as a dictionary 
    where the key hierarchy has been flattened to one level:
    
        {a: 1, b: {ba: 2, bb: {bba: 3}, bc: 4}} becomes:
        {a: 1, b.ba: 2, b.bb.bba: 3, b.bc: 4}
        
    this method corrects the settings hierarchy and returns it to a normal dict 
    structure.
    
    :param given_settings: the settings returned by ES
    :return: the settings as a proper dictionary
    :rtype: dict
    """

    def recursive_defaultdict():
        """
        Helper "class" that makes creating dictionaries of arbitrary size easy
        Each level of the defaultdict defaults to a new recursive_defaultdict.
        """
        return defaultdict(recursive_defaultdict)

    fake_settings = recursive_defaultdict()

    for key, value in given_settings["settings"].items():
        target = fake_settings
        parts = key.split('.')
        for subkey in parts[:-1]:
            target = target[subkey]
        target[parts[-1]] = value

    return convert_numeric_dicts_to_lists(fake_settings)


def convert_numeric_dicts_to_lists(settings):
    """
    Cleans up the dict generated by convert_dotted_keys_to_nested_dictionary by
    converting dicts with all numeric keys to lists so that they will match (==)
    the settings configured in source.
    
    :param settings: settings dictionary to clean 
    :return: corrected settings
    :rtype: dict
    """
    if not isinstance(settings, dict):
        return settings

    if settings and all(key.isdigit() for key in settings.keys()):
        # Found an all-digit map.  Most likely supposed to be a list
        return map(lambda x: x[1], sorted(settings.items(), key=lambda x: int(x[0])))

    for key, value in settings.items():
        settings[key] = convert_numeric_dicts_to_lists(value)
    return settings


def deep_compare_settings(report, index, expected_settings, actual_settings, prefix=""):
    """
    performs a deep comparison of the expected settings against the settings 
    reported by elasticsearch.
    
    :param index:  index settings 
    :param expected_settings: settings as loaded from source 
    :param actual_settings: settings as reported by ES
    """

    for setting, value in expected_settings.items():
        if setting not in actual_settings:
            logger.info("index '{0}' is missing setting '{1}' (should be {2}) "
                        .format(index["index_name"], setting, json.dumps(value, sort_keys=True)))
            report.add_missing_setting(setting, value)
            continue

        # Allow overriding expected settings from the runtime index settings
        # This accommodates for differences in number_of_shards and number_of_replicas
        if setting in index:
            expected_settings[setting] = value = index[setting]

        # Compare as strings because ints may be configured/returned as either strings or ints
        if value == actual_settings[setting] or str(actual_settings[setting]) == str(value):
            continue

        new_prefix = prefix + "." + setting if prefix else setting
        if isinstance(actual_settings[setting], dict) and isinstance(value, dict):
            deep_compare_settings(report, index, value, actual_settings[setting], new_prefix)
        else:
            logger.info("Detected different settings for {0}".format(setting))
            logger.info("Expected Setting:\n" + json.dumps(value, sort_keys=True, indent=2))
            logger.info("Actual Setting:\n" + json.dumps(actual_settings[setting], sort_keys=True, indent=2))
            report.add_different_setting(new_prefix, value, actual_settings[setting])

    for setting, value in actual_settings.items():
        if setting not in expected_settings:
            report.add_extra_setting(setting, value)
    return report


def loosen_type_expectations(expected, actual, strict):
    """
    Contains the "fudges" that exist between mapping definitions in source and 
    what will happen in ES.
    
    Rather than change every mapping to reflect the discrepancies experienced 
    here we just loosen the restrictions to clean up the properties that may be 
    problematic.
    
    :param expected: expected property from source 
    :param actual: actual property from ES
    """

    # Prepare for es 2.x upgrade where analyzer is the same as index_analyzer
    if "index_analyzer" in expected and "search_analyzer" in expected and "analyzer" in actual and "index_analyzer" not in actual and "search_analyzer" not in actual:
        if expected["index_analyzer"] == expected["search_analyzer"] == actual["analyzer"]:
            analyzer = actual["analyzer"]
            expected["analyzer"] = analyzer
            actual["index_analyzer"] = analyzer
            actual["search_analyzer"] = analyzer

    if expected.get("type", None) in ("long", "ip", "boolean"):
        if actual.get("type", None) == expected.get("type", None):
            # "not_analyzed" is not actually a supported value for numeric types
            # https://www.elastic.co/guide/en/elasticsearch/reference/1.7/mapping-core-types.html#number
            if expected.get("index", None) == "not_analyzed" and "index" not in actual:
                actual["index"] = expected["index"]


    # These shouldn't be necessary but a "type": "String" was introduced in one
    # mappings file -_-
    if "type" in actual:
        actual["type"] = actual["type"].lower()

    if "type" in expected:
        expected["type"] = expected["type"].lower()

    # "enabled": false has no effect on non-object fields.
    # The fact that it's present is a bug as we likely intended to set "index": "not_analyzed"
    if expected.get("type", None) != "object" and expected.get("enabled", None) is not None and "enabled" not in actual:
        actual["enabled"] = expected["enabled"]

    # Properties configured as "type": "object" do not show "type": "object"
    # when queried, but contain otherwise identical configurations.
    if expected.get("type", None) == "object" and actual.get("type", None) is None:
        actual["type"] = expected["type"]

    # object type will populate its own properties and so if they're not
    # configured in the expected mappings that's okay
    if expected.get("type", None) == "object" and actual.get("type", None) == "object":
        remove_keys = []
        for prop in actual:
            if prop not in expected:
                remove_keys.append(prop)

        for prop in remove_keys:
            del actual[prop]

    if expected.get("type", None) == "date" and actual.get("type", None) == "date":
        # dateOptionalTime is the default date format
        # https://www.elastic.co/guide/en/elasticsearch/reference/1.7/mapping-core-types.html#date
        # if a date format is specified to be dateOptionalTime by ES but is not
        # configured in source we can safely mark the configuration as the same
        if "format" in actual and "format" not in expected and actual["format"] == "dateOptionalTime":
            expected["format"] = actual["format"]

    if not strict:
        if expected.get("type", None) == "object" and actual.get("type", None) == "object":
            for key in list(expected.keys()):
                if key not in actual:
                    del expected[key]
        if expected.get("type", None) == "string" and actual.get("type", None) == "string":
            if "fielddata" in expected and "fielddata" not in actual:
                del expected["fielddata"]

            if "index" in expected and actual.get("index", None) is None:
                del expected["index"]

        if "fields" in expected and "fields" in actual:
            for subfield in expected["fields"]:
                if subfield in actual["fields"]:
                    loosen_type_expectations(expected["fields"][subfield], actual["fields"][subfield], strict)


def is_same_multi_field(name, expected, actual, allow_unexpected_properties=False):
    """
    compares two fields to see if they are equivalent multi_field specifications
    
    Needed due to potential discrepancies in how multifields are represented.
    
    The following field declarations are equivalent:
    
    "test": {
      "fields": {
        "test": {
          "type": "string"
        },
        "raw": {
          "type": "string":
          "index": "not_analyzed"
        }
      },
      "type": "multi_field"
    }
    
    "test": {
      "type": "string",
      "fields": {
        "raw": {
          "type": "string",
          "index": "not_analyzed"
        }
      }
    }
    
    The first definition contains a child field with the same name as the parent field which is now redundant
    The second definition removes this child field, moving all attributes to the top level
     
    This may or may not be automatically changed, depending on how/when the indices were created.
    
    :param name: The name of the field to check 
    :param expected: Expected field conf (from source)
    :param actual: Actual field conf (from ES)
    :param allow_unexpected_properties: whether additional properties in the actual field should 
    cause validation to fail  
    :return: whether the fields are equivalent
    :rtype: bool
    """

    if "fields" not in expected or "fields" not in actual:
        return False

    for field_name, field in expected["fields"].items():
        if actual["fields"].get(field_name, None) != field:
            logger.info("'{0}' has differing sub-field '{1}'".format(name, field_name))
            logger.info("Expected Field:\n" + json.dumps(field, sort_keys=True, indent=2))
            logger.info("Actual Field:\n" + json.dumps(actual["fields"].get(field_name, None), sort_keys=True, indent=2))
            return False

    for prop, value in expected.items():
        if prop == "fields":
            continue

        if name in actual["fields"]:
            if actual["fields"][name].get(prop, None) != value:
                logger.info("'{0}' has differing properties for parent property '{1}'".format(name, prop))
                logger.info("Expected Property:\n" + json.dumps(value, sort_keys=True, indent=2))
                logger.info("Actual Property:\n" + json.dumps(actual["fields"][name].get(prop, None), sort_keys=True, indent=2))
                return False
        else:
            if actual.get(prop, None) != value:
                logger.info("'{0}' has differing properties for parent property '{1}'".format(name, prop))
                logger.info("Expected Property:\n" + json.dumps(value, sort_keys=True, indent=2))
                logger.info("Actual Property:\n" + json.dumps(actual.get(prop, None), sort_keys=True, indent=2))
                return False

    if name in actual["fields"]:
        for prop, value in actual["fields"][name].items():
            if prop not in expected:
                logger.info("'{0}' has unexpected additional property '{1}' (value {2})".format(name, prop, value))
                if not allow_unexpected_properties:
                    return False

    return True


def deep_compare_mappings(report, index, expected_props, actual_props):
    """
    performs a deep comparison of the expected mappings against the mappings
    reported by elasticsearch.
    
    :param index: index settings
    :param expected_props: Mapping properties as configured in source
    :param actual_props: Mapping properties as reported by ES
    """
    for property in expected_props:
        if property not in actual_props:
            logger.info("index '{index_name}' missing property '{property}'".format(property=property, **index))
            report.add_missing_field(property, expected_props[property])
            continue

        if expected_props[property] == actual_props[property]:
            continue

        expected_prop = expected_props[property]
        actual_prop = actual_props.get(property, {})

        if "properties" in expected_prop:
            sub_state = deep_compare_mappings(IndexReport(parent=report), index, expected_prop["properties"], actual_prop["properties"])
            if sub_state.get_mappings_state() != IndexReport.UP_TO_DATE_MAPPINGS:
                report.add_different_field(property, expected_prop, actual_prop)
            continue

        loosen_type_expectations(expected_prop, actual_prop, report.strict_mappings_validation)
        if "fields" in expected_prop:
            multi_fields_to_remove = []
            for field_name, field in expected_prop["fields"].items():
                actual_subfield = actual_prop.get("fields", {}).get(field_name, {})
                loosen_type_expectations(field, actual_subfield, report.strict_mappings_validation)
                if field == actual_subfield:
                    multi_fields_to_remove.append(field_name)

            # We know these are the same and can just remove them from the comparison at this point
            # for field in multi_fields_to_remove:
            #     del expected_prop["fields"][field]
            #     del actual_prop["fields"][field]
        if expected_prop == actual_prop:
            continue
        elif is_same_multi_field(property, expected_prop, actual_prop, report.strict_mappings_validation):
            continue
        else:
            logger.info("index '{index_name}' has differing property for '{property}'".format(property=property, **index))
            logger.info("Expected Property:\n" + json.dumps(expected_prop, sort_keys=True, indent=2))
            logger.info("Actual Property:\n" + json.dumps(actual_prop, sort_keys=True, indent=2))
            report.add_different_field(property, expected_prop, actual_prop)

    for prop in actual_props:
        if prop not in expected_props:
            logger.info("index '{index_name}' has unexpected property {property}".format(property=prop, **index))
            report.add_extra_field(prop, actual_props[prop])

    return report


def analyze_es_schema(args, settings, auth):
    """
    Analzyes the entire elasticsearch schema to find and highlight differences 
    in schema in source and the schema reported by elasticsearch.
    
    :param settings: expected ES schema definition
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :return: dictionary of indices, indexed by status:
     {MAPPINGS_UP_TO_DATE: [], SETTINGS_UP_TO_DATE: [], MISSING: [], IndexReport.DIFFERENT_MAPPINGS: [], IndexReport.DIFFERENT_SETTINGS: []}
    """
    indices = defaultdict(list)
    for index in settings["schema"]:
        report = IndexReport(index=index,
                             strict_mappings_validation=args.verify_mappings_strict,
                             strict_settings_validation=args.verify_settings_strict)
        if index["is_template"]:
            verify_template(args, report, auth)
        else:
            verify_index(args, report, auth)
        indices[report.get_mappings_state()].append(report)
        indices[report.get_settings_state()].append(report)
        indices["all"].append(report)
        logger.info(str(report))

    return indices