#!/usr/bin/env python

import argparse
import os
import sys
import logging
import json
import yaml
import requests
from requests_aws4auth import AWS4Auth

logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s', level=logging.DEBUG)
logger = logging.getLogger("provision_schema")
MAPPINGS_DIR = os.path.join(os.path.dirname(__file__), 'mappings')
SETTINGS_DIR = os.path.join(os.path.dirname(__file__), 'settings')
DEFAULT_ES_SETTINGS = os.path.join(os.path.dirname(__file__), 'default_configuration.yml')


def load_es_settings(settings_path):
    """
    Loads an elasticsearch settings yaml file
    
    :param settings_path: file to read 
    :return: settings dictionary
    """
    with open(settings_path) as document:
        settings = yaml.safe_load(document.read())
    return settings


def load_es_schema(default_conf, mappings_dir=MAPPINGS_DIR, settings_dir=SETTINGS_DIR, es_url=None, es_event_url=None):
    """
    Loads and merges default and override configurations.
    
    default_conf provides the base index configurations but any field can be 
    overridden by specifying it in the override conf.  If the override_conf 
    contains the entire set of configuration items default_conf can be ommitted
    
    :param default_conf: path to base configuration (not null) 
    :param mappings_dir: custom path to mappings
    :param settings_dir: custom path to settings
    :param es_url: override the es_url in the configuration for all indices
    :param es_event_url: override the es_url for indices on the event cluster
    :return: expected elasticsearch settings
    """
    default_settings = load_es_settings(default_conf)

    for index in default_settings["schema"]:
        # load the mappings/settings files here so we know they are available throughout
        # create_* expects these to be present

        if es_event_url and index["index_name"].startswith('deviceevent'):
            index["es_url"] = format_url(es_event_url)
        elif es_url:
            index["es_url"] = format_url(es_url)
        else:
            index["es_url"] = format_url(index["es_url"])

        if "mappings_file" in index:
            index["mappings"] = load_json_file(mappings_dir, index["mappings_file"])

        if "settings_file" in index:
            index["settings"] = load_json_file(settings_dir, index["settings_file"])
        else:
            index["settings"] = load_json_file(settings_dir, "config_settings.json")

        # yaml override of index sizing for easier env customization
        if "number_of_replicas" in index:
            index["settings"]["number_of_replicas"] = index["number_of_replicas"]
        if "number_of_shards" in index:
            index["settings"]["number_of_shards"] = index["number_of_shards"]

    return default_settings


def load_json_file(directory, filename):
    """
    Loads a json file and returns its content
    :param directory: Directory from which the file will be loaded
    :param filename: Filename to load
    :return: parsed content of the file
    """
    path = os.path.join(directory, filename)
    try:
        data = json.load(open(path))
        logger.debug("Loaded {0}".format(path))
    except ValueError as err:
        logger.error("{0} contains invalid json".format(path), err)
        sys.exit(1)

    return data


def create_alias(index, auth, delete=True):
    """
    Creates an alias for an index in elasticsearch.
    Deletes the alias if it already exists.
    The alias will be named a_{index_name}
    
    :param index: index properties dictionary 
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :param delete: whether the template should be deleted before creating it
    """
    index_name = index["index_name"]

    create_data = {
        "actions": [
            {"add": {"index": index_name, "alias": "a_{0}".format(index_name)}}
        ]
    }
    create_url = "{0}/_aliases".format(index["es_url"])
    delete_url = "{0}/a_{1}".format(index["es_url"], index["index_name"])

    if delete:
        logger.info("Dropping alias a_{0}".format(index_name))
        resp = requests.delete(delete_url, auth=auth, verify=False)
        if resp.status_code not in (200, 404):
            logger.error(resp.json())
            resp.raise_for_status()

    logger.info("Creating alias a_{0}".format(index_name))
    resp = requests.post(create_url, auth=auth, verify=False, data=json.dumps(create_data), headers={'Content-Type': 'application/json'})
    if resp.status_code == 404 and resp.json().get("error", "") == "IndexMissingException[[{index_name}] missing]".format(**index):
        logger.warning("Index {index_name} does not exist... cannot create alias".format(**index))
        return

    if resp.status_code != 200:
        logger.error(resp.json())

    resp.raise_for_status()


def create_index(index, auth, delete=True):
    """
    Creates an index in elasticsearch
    Deletes the index if it already exists
    
    :param index: index properties dictionary
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :param delete: whether the index should be dropped before creating it
    """
    index_name = index["index_name"]
    url = "{0}/{1}".format(index["es_url"], index_name)

    body = {}

    if "mappings" in index:
        body["mappings"] = index["mappings"]
    if "settings" in index:
        body["settings"] = index["settings"]

    if delete:
        logger.info("Dropping {0} index".format(index_name))
        resp = requests.delete(url, auth=auth, verify=False)
        if resp.status_code not in (200, 404):
            resp.raise_for_status()
    logger.info("Creating {0} index".format(index_name))
    resp = requests.put(url, auth=auth, verify=False, data=json.dumps(body), headers={'Content-Type': 'application/json'})
    if resp.status_code not in (200, 404):
        logger.error(resp.json())
        resp.raise_for_status()


def create_template(template, auth, delete=True):
    """
    Creates a template in elasticsearch
    :param template: template parameters
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :param delete: whether the template should be deleted before creating it
    """
    index_name = template["index_name"]
    url = "{0}/_template/template-{1}".format(template["es_url"], index_name)

    body = {
        "template": template.get("template_pattern", "{0}-*".format(index_name)),
        "order": template.get("template_order", 0)
    }

    if "mappings" in template:
        body["mappings"] = template["mappings"]

    if "existing_settings" in template:
        body["settings"] = template["existing_settings"]
    elif "settings" in template:
        body["settings"] = template["settings"]

    if delete:
        # Required to delete data (deleting just the template is insufficient)
        logger.info("Dropping {0} index data".format(index_name))
        requests.delete("{0}/{1}*".format(template["es_url"], index_name), auth=auth, verify=False)

        logger.info("Dropping {0} template".format(index_name))
        resp = requests.delete(url, auth=auth, verify=False)
        if resp.status_code not in (200, 404):
            logger.error(resp.json())
            resp.raise_for_status()

    logger.info("Creating {0} template".format(index_name))
    resp = requests.put(url, auth=auth, verify=False, data=json.dumps(body), headers={'Content-Type': 'application/json'})
    if resp.status_code != 200:
        logger.error(resp.json())
    resp.raise_for_status()


def provision_clean(auth, es_settings, restore_mode):
    """
    Provisions an elasticsearch database from by bringing it into a clean state
    with all relevant indexes, templates, and aliases initialized.
    
    :param auth: AWS Auth (if targeting Amazon Elasticsearch Service)
    :param es_settings: dict of settings for the indexes, aliases, and templates to create
    :param restore_mode: whether the provisioning should be done as a restore or a completely 
    fresh installation.  
    restores (True) will only create aliases and templates.  
    normal (False) will drop then re-create everything
    """
    for index in es_settings["schema"]:
        if index["is_template"]:
            create_template(index, auth, delete=not restore_mode)
        else:
            if not restore_mode:
                create_index(index, auth)
            create_alias(index, auth, delete=not restore_mode)


def format_url(url):
    url = url.rstrip("/")

    if not url.startswith("http"):
        url = "http://{0}".format(url)
    return url


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", dest="default_conf", default=DEFAULT_ES_SETTINGS,
                        help="Location of the base default_configuration.yml file common to all environments")
    parser.add_argument("--alias-and-templates-only", dest="restore_mode", required=False, action='store_true',
                        help="Create ONLY the aliases and templates that are missing. "
                             "Useful in conjunction with elasticdump restores")
    parser.add_argument('--no-confirm', action='store_false', default=True, required=False, dest="confirm",
                        help='Do not prompt before running the schema creation.  '
                             'NOTE:  THIS WILL DELETE CURRENT INDEXES/TEMPLATES')
    parser.add_argument('--es-url-override', type=str, required=False, dest="es_url",
                        help="Globally replace the es_url for elasticsearch with a new one")
    parser.add_argument('--es-event-url-override', type=str, required=False, dest="es_eventurl",
                        help="Globally replace the es_url for event templates with a new one, will override\n"
                             "--es-url-override on the event template loads")
    parser.add_argument("--mappings-dir", default=MAPPINGS_DIR, help="Override the default mappings directory")
    parser.add_argument("--settings-dir", default=SETTINGS_DIR, help="Override the default settings directory")
    parser.add_argument("--aws", action="store_true", help="Targeting an Amazon Elasticsearch Service cluster")
    parser.add_argument("--aws-region", default="us-west-2", help="The base URL for elasticsearch")


    args = parser.parse_args()

    es_settings = load_es_schema(args.default_conf, args.mappings_dir, args.settings_dir, args.es_url, args.es_eventurl)

    if args.confirm:
        logger.warning("Are you sure you want to provision a new ES database?")
        logger.warning("** Current indexes, templates, and aliases will be deleted and recreated **")
        logger.warning("Confirm [y/N]")
        confirm = raw_input()
        if not confirm.strip().startswith("y"):
            logger.error("Aborting provisioning")
            sys.exit(1)

    auth = None
    provision_clean(auth, es_settings, args.restore_mode)


if __name__ == "__main__":
    main()
