#!/usr/bin/env python
import argparse
import json
import logging
import sys
import os
import copy
import requests
import shutil
import urllib3
from requests_aws4auth import AWS4Auth

from es import IndexReport, analyze_es_schema
from provision_elasticsearch import create_index, create_alias, create_template, load_es_schema, \
    MAPPINGS_DIR, SETTINGS_DIR, DEFAULT_ES_SETTINGS

logging.basicConfig(format='%(asctime)s [%(levelname)s] %(message)s', level=logging.WARNING, stream=sys.stdout)
logging.getLogger("provision_schema").setLevel(logging.INFO)
# suppress http request logging
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("verify_schema").setLevel(logging.WARNING)
logger = logging.getLogger("upgrade_schema")
logger.setLevel(logging.WARNING)


urllib3.disable_warnings()


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", dest="default_conf", default=DEFAULT_ES_SETTINGS,
                        help="Location of the base default_configuration.yml file common to all environments")

    parser.add_argument("--mappings-dir", default=MAPPINGS_DIR, help="Override the default mappings directory")
    parser.add_argument("--settings-dir", default=SETTINGS_DIR, help="Override the default settings directory")

    parser.add_argument('--es-url-override', type=str, required=False, dest="es_url",
                        help="Globally replace the es_url for elasticsearch with a new one")
    parser.add_argument('--es-event-url-override', type=str, required=False, dest="es_eventurl",
                        help="Globally replace the es_url for event templates with a new one, will override\n"
                             "--es-url-override on the event template loads")
    parser.add_argument("--es-user", default=os.environ.get('ES_USER'), help="User for elasticsearch")
    parser.add_argument("--es-password", default=os.environ.get('ES_PASSWORD'), help="Password for elasticsearch")

    parser.add_argument("--mappings", action="store_true", required=False, default=False, dest="verify_mappings",
                        help="Verify the mappings of the indices/templates. "
                             "Only breaking changes reported.")
    parser.add_argument("--mappings-strict", action="store_true", required=False, default=False, dest="verify_mappings_strict",
                        help="Strictly verify the mappings for indices/templates. "
                             "All differences reported.")
    parser.add_argument("--settings", action="store_true", required=False, default=False, dest="verify_settings",
                        help="Verify the settings of the indices/templates. ")
    parser.add_argument("--settings-strict", action="store_true", required=False, default=False, dest="verify_settings_strict",
                        help="Strictly verify the settings for indices/templates. "
                             "All differences reported.")
    parser.add_argument("-v", "--verbose", action="count", required=False, default=0, dest="verbose",
                        help="Be verbose.  Can be repeated. "
                             "One verbose shows minified differences. "
                             "Two verbose shows pretty printed differences. "
                             "Three verbose shows full debug output during testing.")
    parser.add_argument("--continue-on-error", action="store_true", required=False, default=False, dest="continue_on_error",
                        help="Continue adding missing fields if one field fails to be added")

    parser.add_argument("--add-missing-indices", action="store_true", required=False, default=False,
                        help="Automatically add indices/templates that are missing")
    parser.add_argument("--add-missing-fields", action="store_true", required=False, default=False,
                        help="Automatically add fields to indices that are missing them")
    parser.add_argument("--correct-templates", action="store_true", required=False, default=False,
                        help="Automatically update differing templates to the latest from source."
                             "Does not correct the indices created from the bad templates, but will "
                             "ensure new indices are created correctly.")

    parser.add_argument("--correct-fields", action="store_true", required=False, default=False,
                        help="Automatically update differing fields to the latest from source."
                             "This is a best-effort approach based on the put mapping api.  "
                             "It doesn't hurt to try to fix all the mappings we can")

    parser.add_argument("--no-confirm", action="store_true", required=False, default=False,
                        help="Disable confirmation before taking action (dangerous!)")

    parser.add_argument("--output-dir", required=False,
                        help="Provide a directory to store the current mappings and settings for an index")

    args = parser.parse_args()

    if args.verify_mappings_strict:
        args.verify_mappings = True

    if args.verify_settings_strict:
        args.verify_settings = True

    return args


def confirm(target, operation):
    """
    Helper function to show a confirmation message
    
    :param target: targets of the operation needing confirmation 
    :param operation: string message to show before the prompt
    :return: True, if confirmed. False otherwise
    """
    if not target:
        return False

    logger.warning("Are you sure you want to {0}?".format(operation))
    logger.warning(", ".join([t.index_name for t in target]))
    logger.warning("Confirm [y/N]")
    confirm = raw_input()
    if not confirm.strip().startswith("y"):
        logger.error("Aborting")
        return False

    return True


def create_new_indices(indices_to_create, no_confirm, auth):
    """
    Creates missing indices/templates
    
    :param indices_to_create: list of IndexReports that need to be created 
    :param no_confirm: whether the creation should proceed automatically or require user confirmation
    :param auth: Basic Auth tuple (if targeting Amazon Elasticsearch Service)
    """

    if no_confirm or confirm(indices_to_create, "create these new indices"):
        for report in indices_to_create:
            index = report.index
            if index["is_template"]:
                logger.info("Creating template {index_name} on {es_url}".format(**index))
                create_template(index, auth, delete=False)
            else:
                logger.info("Creating index {index_name} on {es_url}".format(**index))
                create_index(index, auth, delete=False)
                create_alias(index, auth, delete=False)


def correct_templates(templates_to_update, no_confirm, auth):
    """
    Updates existing templates that have differing configurations from those in source
    
    :param templates_to_update: list of IndexReports that need to be created.  Filtering to only templates occurs here
    :param no_confirm: whether the creation should proceed automatically or require user confirmation
    :param auth: Basic Auth tuple (if targeting Amazon Elasticsearch Service)
    """

    templates_to_update = [t for t in templates_to_update if t.index["is_template"]]

    if no_confirm or confirm(templates_to_update, "update these templates to what is in source"):
        for report in templates_to_update:
            index = report.index
            index["existing_settings"] = report.actual_settings
            create_template(index, auth, delete=False)


def add_missing_fields(indices_to_update, no_confirm, continue_on_error, auth):
    """
    Adding new fields is free!  Let's do it!
    
    https://www.elastic.co/blog/changing-mapping-with-zero-downtime
    :param indices_to_update: List of indices (this method checks for missing fields in the report - no pre-filtering necessary
    :param no_confirm: Whether to ask via stdin before adding fields
    :param continue_on_error: Whether errors adding a field should stop processing remaining fields
    :param auth: Basic Auth tuple (if targeting Amazon Elasticsearch Service)
    """
    indices_with_missing_fields = []
    for report in indices_to_update:

        if report.missing_fields and not report.is_template:
            indices_with_missing_fields.append(report)
        for child in report.children:
            if child.missing_fields and not child.is_template:
                indices_with_missing_fields.append(child)

    if no_confirm or confirm(indices_with_missing_fields, "add missing fields to these indices"):
        for report in indices_with_missing_fields:
            name = list(report.index["mappings"].keys())[0]
            logging.warning("  ** " + report.index_name)

            for (field_name, risk, configuration) in report.missing_fields:
                logging.warning("      " + field_name)

                # add these new properties to the {name} type of {index}
                # https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-put-mapping.html
                url = "{es_url}/{index}/_mapping/{name}".format(index=report.index_name, name=name, **report.index)
                if not report.index["is_template"]:
                    url = "{es_url}/{index}/_mapping/{name}".format(index="a_" + report.index_name, name=name, **report.index)
                request_body = {
                    name: {
                        "properties": {
                            field_name: configuration
                        }
                    }
                }
                resp = requests.put(url, json=request_body, auth=auth, verify=False)
                if resp.status_code != 200:
                    logger.error("         ** Error adding field!")
                    logger.error(resp.content)
                    logger.error("         ** Attempted to apply {}".format(json.dumps(request_body)))
                    if not continue_on_error:
                        resp.raise_for_status()


def try_replace_mappings_differing_fields(indices_to_update, no_confirm, auth):
    """
    Why should we try to figure out if a mapping update will work.  Let's just let ES fail it if it isn't allowed
    
    https://www.elastic.co/guide/en/elasticsearch/reference/1.7/indices-put-mapping.html#merging-conflicts
    :param indices_to_update: 
    :param no_confirm:
    :param auth: Basic Auth tuple (if targeting Amazon Elasticsearch Service)
    """
    indices_with_different_fields = []
    for report in indices_to_update:
        if report.different_fields and not report.is_template:
            indices_with_different_fields.append(report)
        for child in report.children:
            if child.different_fields and not child.is_template:
                indices_with_different_fields.append(child)

    if no_confirm or confirm(indices_with_different_fields, " put updated mappings for these indices"):
        for report in indices_with_different_fields:
            name = list(report.index["mappings"].keys())[0]
            logging.warning("  ** " + report.index_name)
            url = "{es_url}/{index}/_mapping/{name}".format(index=report.index_name, name=name, **report.index)
            if not report.index["is_template"]:
               url = "{es_url}/{index}/_mapping/{name}".format(index="a_"+report.index_name, name=name, **report.index)
            logging.warning("  ** URL **  " + url)
            for (field_name, risk, configuration, actual) in report.different_fields:
                logging.warning("      " + field_name)

                if "fields" in configuration and "fields" in actual and not actual["fields"]:
                    # put the subfields first since these are definitely safe
                    add_fields = copy.deepcopy(actual)
                    add_fields["fields"] = configuration["fields"]
                    put_mappings_for_field(add_fields, field_name, name, url, auth)

                put_mappings_for_field(configuration, field_name, name, url, auth)


def put_mappings_for_field(configuration, field_name, name, url, auth):
    request_body = {
        name: {
            "properties": {
                field_name: configuration
            }
        }
    }

    # instead of batching we will perform updates in the smallest possible units to
    # have the greatest chance for success
    resp = requests.put(url, json=request_body, auth=auth, verify=False)
    if resp.status_code != 200:
        if "MergeMappingException" in resp.text:
            logger.error("          could not merge source mappings with current mappings for " + field_name)
        logger.error(resp.text)


def print_analysis(args, indices):
    """
    Logs out the final report of indices that need attention
    :param indices: dict of index statuses
    """
    logger.warning("-" * 70)
    logger.warning("ElasticSearch Schema Report")

    for report_type, reports in indices.items():
        logger.warning(f"    {report_type}: {len(reports)}")

    if indices[IndexReport.MISSING_MAPPINGS]:
        logger.warning("Discovered {0} missing indices/templates:".format(len(indices[IndexReport.MISSING_MAPPINGS])))
        for index in indices[IndexReport.MISSING_MAPPINGS]:
            logger.warning("  * {index_name}".format(index_name=index.index_name))

    # The templates themselves might be okay, but their derived indices may be out of date so we
    # need to check all indices here.
    indices_with_different_mappings = [index for index in indices["all"] if
                                       index.get_mappings_state() == IndexReport.DIFFERENT_MAPPINGS or any(
                                           child.get_mappings_state() == IndexReport.DIFFERENT_MAPPINGS for child in
                                           index.children)]
    if indices_with_different_mappings:
        logger.warning(
            "Discovered {0} indices/templates with incorrect mappings:".format(len(indices_with_different_mappings)))
        for index in indices_with_different_mappings:
            logger.warning("  * {index_name}".format(index_name=index.index_name))
            for difference in index.get_mappings_differences(verbosity=args.verbose):
                logger.warning("      " + difference)

    if indices[IndexReport.DIFFERENT_SETTINGS]:
        logger.warning(
            "Discovered {0} indices/templates with incorrect settings:".format(len(indices[IndexReport.DIFFERENT_SETTINGS])))
        for index in indices[IndexReport.DIFFERENT_SETTINGS]:
            logger.warning("  * {index_name}".format(index_name=index.index_name))
            for difference in index.get_settings_differences(verbosity=args.verbose):
                logger.warning("      " + difference)

    if not indices[IndexReport.MISSING_MAPPINGS] and not indices[IndexReport.DIFFERENT_MAPPINGS] and not indices[IndexReport.DIFFERENT_SETTINGS]:
        logger.warning("Schema up to date!")

    logger.warning("-" * 70)


def save_json_file(directory, filename, data):
    """
    Writes out a pretty-printed json payload to a given file
    :param directory: Directory to which the file will be write
    :param filename: Filename to write
    :param data: Data to write
    """


    path = os.path.join(directory, filename)
    with open(path, "w") as output:
        json.dump(data, output, indent=2, sort_keys=True)


def save_mappings_and_settings(output_dir, indices):
    """
    Persists a snapshot of the mappings and settings found in the actual ES instance
    :param output_dir: root directory for outputting these configurations 
    :param indices: list of indices to persist
    """
    try:
        shutil.rmtree(output_dir)
    except OSError as e:
        # errno 2 is "No such file or directory"... expected
        if e.errno != 2:
            raise e
    mappings_output = os.path.join(output_dir, 'mappings')
    settings_output = os.path.join(output_dir, 'settings')

    os.mkdir(output_dir)
    os.mkdir(mappings_output)
    os.mkdir(settings_output)

    for report in indices["all"]:
        if report.actual_mappings:
            save_json_file(mappings_output, report.index["mappings_file"], report.actual_mappings)

        if report.actual_settings:
            save_json_file(settings_output, report.index["settings_file"], report.actual_settings)

        for child in report.children:
            if child.actual_mappings:
                save_json_file(mappings_output, child.index_name + "__" + child.index["mappings_file"], child.actual_mappings)

            if child.actual_settings:
                save_json_file(settings_output, child.index_name + "__" + child.index["settings_file"], child.actual_settings)


def main():
    args = parse_args()
    if args.verbose >= 3:
        logger.setLevel(logging.DEBUG)
        logging.getLogger("verify_schema").setLevel(logging.DEBUG)

    auth = None
    if args.es_user and args.es_password:
        auth = (args.es_user, args.es_password)
        
    settings = load_es_schema(args.default_conf, args.mappings_dir, args.settings_dir, args.es_url, args.es_eventurl)
    indices = analyze_es_schema(args, settings, auth)

    print_analysis(args, indices)

    if args.add_missing_indices:
        create_new_indices(indices[IndexReport.MISSING_MAPPINGS], args.no_confirm, auth)

    if args.correct_fields:
        try_replace_mappings_differing_fields(indices["all"], args.no_confirm, auth)

    if args.correct_templates:
        correct_templates(indices[IndexReport.DIFFERENT_MAPPINGS], args.no_confirm, auth)

    if args.add_missing_fields:
        add_missing_fields(indices["all"], args.no_confirm, args.continue_on_error, auth)

    if args.output_dir:
        save_mappings_and_settings(args.output_dir, indices)


if __name__ == '__main__':
    main()
